﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

/// 参考にしたページ
/// ttp://homepage2.nifty.com/nonnon/SoftSample/CS.NET/SampleBitmapPlus.html
/// 参考というかほぼそのまま使った。usingで使えるようにIDisposableを継承した
namespace BitmapPlus
{
	/// <summary>
	/// Bitmap処理を高速化するためのクラス
	/// </summary>
	public sealed class BitmapHighSpeed : IDisposable	// namespace と class の名前が同じだと少し面倒なことになる
	{
		/// <summary>
		/// オリジナルのBitmapオブジェクト
		/// </summary>
		private readonly Bitmap bmp = null;

		/// <summary>
		/// Bitmapに直接アクセスするためのオブジェクト
		/// </summary>
		private BitmapData img = null;

		/// <summary>
		/// コンストラクタ(Bitmap処理の高速化開始)
		/// </summary>
		/// <param name="original">オリジナルのBitmapオブジェクト</param>
		public BitmapHighSpeed(Bitmap original)
		{
			if (original == null)
			{
				throw new ArgumentNullException("original", "引数が null です。");
			}
			// オリジナルのBitmapオブジェクトを保存
			this.bmp = original;
			// Bitmapに直接アクセスするためのオブジェクト取得(LockBits)
			this.img = this.bmp.LockBits(new Rectangle(0, 0, this.bmp.Width, this.bmp.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
		}

		/// <summary>
		/// デストラクタ(Bitmap処理の高速化終了)
		/// </summary>
		~BitmapHighSpeed()
		{
			this.Dispose();
		}

		/// <summary>
		/// 開放(Bitmap処理の高速化終了)
		/// </summary>
		public void Dispose()
		{
			if (this.img != null)
			{
				// Bitmapに直接アクセスするためのオブジェクト開放(UnlockBits)
				this.bmp.UnlockBits(this.img);
				this.img = null;
			}
		}

		/*
		/// <summary>
		/// Bitmap処理の高速化開始
		/// </summary>
		public void BeginAccess()
		{
			// Bitmapに直接アクセスするためのオブジェクト取得(LockBits)
			_img = _bmp.LockBits(new Rectangle(0, 0, _bmp.Width, _bmp.Height),
				System.Drawing.Imaging.ImageLockMode.ReadWrite,
				System.Drawing.Imaging.PixelFormat.Format24bppRgb);
		}

		/// <summary>
		/// Bitmap処理の高速化終了
		/// </summary>
		public void EndAccess()
		{
			if (_img != null)
			{
				// Bitmapに直接アクセスするためのオブジェクト開放(UnlockBits)
				_bmp.UnlockBits(_img);
				_img = null;
			}
		}
		*/

		/// <summary>
		/// BitmapのGetPixel同等
		/// </summary>
		/// <param name="x">Ｘ座標</param>
		/// <param name="y">Ｙ座標</param>
		/// <returns>Colorオブジェクト</returns>
		public Color GetPixel(int x, int y)
		{
			if (this.img == null)
			{
				// 既にDisposeが呼び出されている場合はBitmap標準のGetPixel
				return this.bmp.GetPixel(x, y);
			}

			// Bitmap処理の高速化を開始している場合はBitmapメモリへの直接アクセス
			IntPtr adr = this.img.Scan0;
			int pos = x * 3 + this.img.Stride * y;
			byte b = Marshal.ReadByte(adr, pos + 0);
			byte g = Marshal.ReadByte(adr, pos + 1);
			byte r = Marshal.ReadByte(adr, pos + 2);
			return Color.FromArgb(r, g, b);
		}

		/// <summary>
		/// このクラスのGetPixelのif文を省略
		/// </summary>
		/// <param name="x">Ｘ座標</param>
		/// <param name="y">Ｙ座標</param>
		/// <returns>Colorオブジェクト</returns>
		public Color GetPixel2(int x, int y)
		{
			// Bitmap処理の高速化を開始している場合はBitmapメモリへの直接アクセス
			IntPtr adr = this.img.Scan0;
			int pos = x * 3 + this.img.Stride * y;
			byte b = Marshal.ReadByte(adr, pos + 0);
			byte g = Marshal.ReadByte(adr, pos + 1);
			byte r = Marshal.ReadByte(adr, pos + 2);
			return Color.FromArgb(r, g, b);
		}

		/// <summary>
		/// BitmapのSetPixel同等
		/// </summary>
		/// <param name="x">Ｘ座標</param>
		/// <param name="y">Ｙ座標</param>
		/// <param name="col">Colorオブジェクト</param>
		public void SetPixel(int x, int y, Color col)
		{
			if (this.img == null)
			{
				// 既にDisposeが呼び出されている場合はBitmap標準のSetPixel
				this.bmp.SetPixel(x, y, col);
				return;
			}

			// Bitmap処理の高速化を開始している場合はBitmapメモリへの直接アクセス
			IntPtr adr = this.img.Scan0;
			int pos = x * 3 + this.img.Stride * y;
			Marshal.WriteByte(adr, pos + 0, col.B);
			Marshal.WriteByte(adr, pos + 1, col.G);
			Marshal.WriteByte(adr, pos + 2, col.R);
		}

		/// <summary>
		/// このクラスのSetPixelのif文を省略
		/// </summary>
		/// <param name="x">Ｘ座標</param>
		/// <param name="y">Ｙ座標</param>
		/// <param name="col">Colorオブジェクト</param>
		public void SetPixel2(int x, int y, Color col)
		{
			// Bitmap処理の高速化を開始している場合はBitmapメモリへの直接アクセス
			IntPtr adr = this.img.Scan0;
			int pos = x * 3 + this.img.Stride * y;
			Marshal.WriteByte(adr, pos + 0, col.B);
			Marshal.WriteByte(adr, pos + 1, col.G);
			Marshal.WriteByte(adr, pos + 2, col.R);
		}
	}
}
