﻿using System;
using System.Collections.Generic;
using System.Drawing;
using BitmapPlus;
using XorShift;

namespace MapGenerator
{
	public sealed class Map
	{
		private readonly Xor128 rnd;
		private readonly uint seedX;
		private readonly uint seedY;
		private readonly uint seedZ;
		private readonly uint seedW;
		private readonly Size size;
		private readonly Grid[,] grids;
		private readonly List<Division> divisionList = new List<Division>();
		private readonly int leftSpace;
		private readonly int topSpace;
		private readonly int rightSpace;
		private readonly int bottomSpace;
		private readonly Size minDivisionSize;
		private readonly Size maxDivisionSize;
		private readonly Size minRoomSize;
		private readonly Size maxRoomSize;
		private readonly int connectionAdditionRate;
		private Bitmap image;

		public uint SeedX { get { return this.seedX; } }
		public uint SeedY { get { return this.seedY; } }
		public uint SeedZ { get { return this.seedZ; } }
		public uint SeedW { get { return this.seedW; } }
		public int Right { get { return this.size.Width - 1; } }
		public int Bottom { get { return this.size.Height - 1; } }
		public int LeftSpace { get { return this.leftSpace; } }
		public int TopSpace { get { return this.topSpace; } }
		public int RightSpace { get { return this.rightSpace; } }
		public int BottomSpace { get { return this.bottomSpace; } }
		public Size MinRoomSize { get { return this.minRoomSize; } }
		public Size MaxRoomSize { get { return this.maxRoomSize; } }

		public Grid this[int x, int y] { get { return this.grids[x, y]; } set { this.grids[x, y] = value; } }
		public Grid this[Point coord] { get { return this.grids[coord.X, coord.Y]; } set { this.grids[coord.X, coord.Y] = value; } }

		private Map(ushort width, ushort height, byte leftSpace, byte topSpace, byte rightSpace, byte bottomSpace, byte minDivisionWidth, byte minDivisionHeight, byte maxDivisionWidth, byte maxDivisionHeight, byte minRoomWidth, byte minRoomHeight, byte maxRoomWidth, byte maxRoomHeight, ushort connectionAdditionRate)
		{
			if (width < 3 + leftSpace + rightSpace)
			{
				throw new ArgumentOutOfRangeException("width", "width が小さすぎます。(3 + leftSpace + rightSpace) 以上の値である必要があります。");
			}
			if (height < 3 + topSpace + bottomSpace)
			{
				throw new ArgumentOutOfRangeException("height", "height が小さすぎます。(3 + topSpace + bottomSpace) 以上の値である必要があります。");
			}
			if (maxDivisionWidth > width - leftSpace - rightSpace)
			{
				throw new ArgumentOutOfRangeException("maxDivisionWidth", "maxDivisionWidth が大きすぎます。(width - leftSpace - rightSpace) 以下の値である必要があります。");
			}
			if (maxDivisionHeight > height - topSpace - bottomSpace)
			{
				throw new ArgumentOutOfRangeException("maxDivisionHeight", "maxDivisionHeight が大きすぎます。(height - topSpace - bottomSpace) 以下の値である必要があります。");
			}
			if (minDivisionWidth > maxDivisionWidth)
			{
				throw new ArgumentOutOfRangeException("minDivisionWidth", "minDivisionWidth が大きすぎます。maxWidth 以下の値である必要があります。");
			}
			if (minDivisionWidth < 3)
			{
				throw new ArgumentOutOfRangeException("minDivisionWidth", "minDivisionWidth が小さすぎます。3 以上の値である必要があります。");
			}
			if (minDivisionHeight > maxDivisionHeight)
			{
				throw new ArgumentOutOfRangeException("minDivisionHeight", "minDivisionHeight が大きすぎます。maxDivisionHeight 以下の値である必要があります。");
			}
			if (minDivisionHeight < 3)
			{
				throw new ArgumentOutOfRangeException("minDivisionHeight", "minDivisionHeight が小さすぎます。3 以上の値である必要があります。");
			}
			if (maxRoomWidth > maxDivisionWidth - 4)
			{
				throw new ArgumentOutOfRangeException("maxRoomWidth", "maxRoomWidth が大きすぎます。(maxDivisionWidth - 4) 以下の値である必要があります。");
			}
			if (maxRoomHeight > maxDivisionHeight - 4)
			{
				throw new ArgumentOutOfRangeException("maxRoomHeight", "maxRoomHeight が大きすぎます。(maxDivisionHeight - 4) 以下の値である必要があります。");
			}
			if (minRoomWidth > maxRoomWidth)
			{
				throw new ArgumentOutOfRangeException("minRoomWidth", "minRoomWidth が大きすぎます。maxRoomWidth 以下の値である必要があります。");
			}
			if (minRoomWidth > minDivisionWidth - 4)
			{
				throw new ArgumentOutOfRangeException("minRoomWidth", "minRoomWidth が大きすぎます。(minDivisionWidth - 4) 以下の値である必要があります。");
			}
			if (minRoomHeight > maxRoomHeight)
			{
				throw new ArgumentOutOfRangeException("minRoomHeight", "minRoomHeight が大きすぎます。maxRoomHeight 以下の値である必要があります。");
			}
			if (minRoomHeight > minDivisionHeight - 4)
			{
				throw new ArgumentOutOfRangeException("minRoomHeight", "minRoomHeight が大きすぎます。(minDivisionHeight - 4) 以下の値である必要があります。");
			}
			this.size = new Size(width, height);
			this.grids = new Grid[this.size.Width, this.size.Height];
			this.leftSpace = leftSpace;
			this.topSpace = topSpace;
			this.rightSpace = rightSpace;
			this.bottomSpace = bottomSpace;
			this.minDivisionSize = new Size(minDivisionWidth, minDivisionHeight);
			this.maxDivisionSize = new Size(maxDivisionWidth, maxDivisionHeight);
			this.minRoomSize = new Size(minRoomWidth, minRoomHeight);
			this.maxRoomSize = new Size(maxRoomWidth, maxRoomHeight);
			this.connectionAdditionRate = connectionAdditionRate;
		}

		/// <summary>
		/// マップ生成クラスコンストラクタ
		/// </summary>
		/// <param name="width">幅(5 ~ 65535)</param>
		/// <param name="height">高さ(5 ~ 65535)</param>
		/// <param name="leftSpace">左端余白(0 ~ 255)</param>
		/// <param name="topSpace">上端余白(0 ~ 255)</param>
		/// <param name="rightSpace">右端余白(0 ~ 255)</param>
		/// <param name="bottomSpace">下端余白(0 ~ 255)</param>
		/// <param name="minDivisionWidth">区画の幅の下限(5 ~ 255)</param>
		/// <param name="minDivisionHeight">区画の高さの下限(5 ~ 255)</param>
		/// <param name="maxDivisionWidth">区画の幅の上限(5 ~ 255)</param>
		/// <param name="maxDivisionHeight">区画の高さの上限(5 ~ 255)</param>
		/// <param name="minRoomWidth">部屋の幅の最小値(1 ~ 251)</param>
		/// <param name="minRoomHeight">部屋の高さの最小値(1 ~ 251)</param>
		/// <param name="maxRoomWidth">部屋の幅の最大値(1 ~ 251)</param>
		/// <param name="maxRoomHeight">部屋の高さの最大値(1 ~ 251)</param>
		/// <param name="connectionAdditionRate">部屋に通路を追加する確率 (value / 1000)</param>
		/// <!--X, Y, Z, W 全てに 0 を入れると、乱数の種は適当に設定される-->
		/// <param name="seedX">乱数の種X</param>
		/// <param name="seedY">乱数の種Y</param>
		/// <param name="seedZ">乱数の種Z</param>
		/// <param name="seedW">乱数の種W</param>
		public Map(ushort width, ushort height, byte leftSpace, byte topSpace, byte rightSpace, byte bottomSpace, byte minDivisionWidth, byte minDivisionHeight, byte maxDivisionWidth, byte maxDivisionHeight, byte minRoomWidth, byte minRoomHeight, byte maxRoomWidth, byte maxRoomHeight, ushort connectionAdditionRate, uint seedX, uint seedY, uint seedZ, uint seedW)
			: this(width, height, leftSpace, topSpace, rightSpace, bottomSpace, minDivisionWidth, minDivisionHeight, maxDivisionWidth, maxDivisionHeight, minRoomWidth, minRoomHeight, maxRoomWidth, maxRoomHeight, connectionAdditionRate)
		{
			if (seedX == 0 && seedY == 0 && seedZ == 0 && seedW == 0)
			{
				this.rnd = new Xor128();
				this.seedX = this.rnd.X;
				this.seedY = this.rnd.Y;
				this.seedZ = this.rnd.Z;
				this.seedW = this.rnd.W;
			}
			else
			{
				this.rnd = new Xor128(seedX, seedY, seedZ, seedW);
				this.seedX = seedX;
				this.seedY = seedY;
				this.seedZ = seedZ;
				this.seedW = seedW;
			}
			// 区画分割
			List<Division> firstDivList = new List<Division>();
			List<Division> nextDivList = new List<Division>();
			firstDivList.Add(new Division(this, this.rnd, this.leftSpace, this.topSpace, this.size.Width - this.rightSpace - this.leftSpace, this.size.Height - this.bottomSpace - this.topSpace));
			while (true)
			{
				int split = 0;
				for (int i = 0; i < firstDivList.Count; ++i)
				{
					Division[] divArray = firstDivList[i].SplitRandom(this.rnd.Next(this.minDivisionSize.Width, this.maxDivisionSize.Width + 1), this.rnd.Next(this.minDivisionSize.Height, this.maxDivisionSize.Height));
					if (divArray == null)
					{
						this.divisionList.Add(firstDivList[i]);
					}
					else
					{
						++split;
						nextDivList.AddRange(divArray);
					}
				}
				if (split == 0)
				{
					break;
				}
				firstDivList = nextDivList;
				nextDivList = new List<Division>();
			}
			// 部屋生成
			for (int i = 0; i < this.divisionList.Count; ++i)
			{
				this.divisionList[i].CreateRoom();
			}
			// 接点をランダムに追加
			if (this.connectionAdditionRate > 999)
			{
				for (int i = 0; i < this.divisionList.Count; ++i)
				{
					this.divisionList[i].ConnectRandom();
				}
			}
			else if (this.connectionAdditionRate > 0)
			{
				for (int i = 0; i < this.divisionList.Count; ++i)
				{
					if (this.rnd.Next(1000) < this.connectionAdditionRate)
					{
						this.divisionList[i].ConnectRandom();
					}
				}
			}
			// 通路生成
			for (int i = 0; i < this.divisionList.Count; ++i)
			{
				this.divisionList[i].CreatePassages();
			}
		}

		public Division[] GetDivisions()
		{
			return this.divisionList.ToArray();
		}

		private void CreateImage()
		{
			this.image = new Bitmap(this.size.Width, this.size.Height);
			using (BitmapHighSpeed imageHS = new BitmapHighSpeed(this.image))
			{
				/// 参考 ttp://d.hatena.ne.jp/Schima/20091118/1258518836
				unsafe
				{
					int rows = this.grids.GetLength(0);
					int cols = this.grids.GetLength(1);
					for (int r = 0; r < rows; ++r)
					{
						fixed (Grid* p = &this.grids[r, 0])
						{
							for (int c = 0; c < cols; ++c)
							{
								Grid grid = p[c];
								if ((grid & Grid.Doorway) != 0)
								{
									imageHS.SetPixel2(r, c, Color.Red);
								}
								else if ((grid & Grid.Room) != 0)
								{
									imageHS.SetPixel2(r, c, Color.White);
								}
								else if ((grid & Grid.Passage) != 0)
								{
									imageHS.SetPixel2(r, c, Color.Gray);
								}
							}
						}
					}
				}
			}
		}

		public Bitmap GetImage()
		{
			if (this.image == null)
			{
				this.CreateImage();
			}
			return this.image;
		}

		public Bitmap GetImage(int mag)
		{
			if (this.image == null)
			{
				this.CreateImage();
			}
			if (mag == 1)
			{
				return this.image;
			}
			else if (mag < 1)
			{
				throw new ArgumentOutOfRangeException("mag", "倍率は 1 以上である必要があります。");
			}
			Bitmap exImage = new Bitmap(this.size.Width * mag, this.size.Height * mag, this.image.PixelFormat);
			using (Graphics g = Graphics.FromImage(exImage))
			{
				g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
				g.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.Half;
				g.DrawImage(this.image, 0, 0, exImage.Width, exImage.Height);
			}
			return exImage;
		}

		public Point GetRandomPos()
		{
			int index = this.rnd.Next(this.divisionList.Count);
			return new Point(this.rnd.Next(this.divisionList[index].Room.Width) + this.divisionList[index].Room.Left, this.rnd.Next(this.divisionList[index].Room.Height) + this.divisionList[index].Room.Top);
		}
	}
}
